@@ -27,15 +27,15 @@
 extern int WIN32_SSL_password_prompt(tcn_pass_cb_t *data);
 #endif
 
-#ifdef HAVE_OCSP_STAPLING
+#ifdef HAVE_OCSP
 #include <openssl/bio.h>
 #include <openssl/ocsp.h>
 /* defines with the values as seen by the asn1parse -dump openssl command */
 #define ASN1_SEQUENCE 0x30
 #define ASN1_OID      0x06
 #define ASN1_STRING   0x86
-static int ssl_verify_OCSP(int ok, X509_STORE_CTX *ctx);
-static int ssl_ocsp_request(X509 *cert, X509 *issuer);
+static int ssl_verify_OCSP(X509_STORE_CTX *ctx, int timeout, int verifyFlags);
+static int ssl_ocsp_request(X509 *cert, X509 *issuer, X509_STORE_CTX *ctx, int timeout, int verifyFlags);
 #endif
 
 /*  _________________________________________________________________
@@ -52,8 +52,9 @@ static int ssl_ocsp_request(X509 *cert, X509 *issuer);
  */
 static int SSL_app_data2_idx = -1;
 static int SSL_app_data3_idx = -1;
+static int SSL_app_data4_idx = -1;
 
-void SSL_init_app_data2_3_idx(void)
+void SSL_init_app_data_idx(void)
 {
     int i;
 
@@ -78,6 +79,15 @@ void SSL_init_app_data2_3_idx(void)
                                  "Third Application Data for SSL",
                                   NULL, NULL, NULL);
 
+    if (SSL_app_data4_idx > -1) {
+        return;
+    }
+
+    SSL_app_data4_idx =
+            SSL_get_ex_new_index(0,
+                                 "Fourth Application Data for SSL",
+                                  NULL, NULL, NULL);
+
 }
 
 void *SSL_get_app_data2(SSL *ssl)
@@ -102,6 +112,16 @@ void SSL_set_app_data3(SSL *ssl, void *arg)
     SSL_set_ex_data(ssl, SSL_app_data3_idx, arg);
 }
 
+void *SSL_get_app_data4(const SSL *ssl)
+{
+    return SSL_get_ex_data(ssl, SSL_app_data4_idx);
+}
+
+void SSL_set_app_data4(SSL *ssl, void *arg)
+{
+    SSL_set_ex_data(ssl, SSL_app_data4_idx, arg);
+}
+
 /* Simple echo password prompting */
 int SSL_password_prompt(tcn_pass_cb_t *data)
 {
@@ -281,7 +301,7 @@ int SSL_CTX_use_certificate_chain(SSL_CTX *ctx, const char *file,
             BIO_free(bio);
             return -1;
         }
-        while (ERR_get_error() > 0) ;
+        while (SSL_ERR_get() > 0) ;
     }
     BIO_free(bio);
     return n;
@@ -292,7 +312,6 @@ int SSL_CTX_use_certificate_chain(SSL_CTX *ctx, const char *file,
  * does client authentication and verifies the certificate chain.
  */
 
-
 int SSL_callback_SSL_verify(int ok, X509_STORE_CTX *ctx)
 {
    /* Get Apache context back through OpenSSL context */
@@ -300,19 +319,27 @@ int SSL_callback_SSL_verify(int ok, X509_STORE_CTX *ctx)
                                           SSL_get_ex_data_X509_STORE_CTX_idx());
     tcn_ssl_conn_t *con = (tcn_ssl_conn_t *)SSL_get_app_data(ssl);
     /* Get verify ingredients */
-    int errnum   = X509_STORE_CTX_get_error(ctx);
-    int errdepth = X509_STORE_CTX_get_error_depth(ctx);
-    int verify   = con->ctx->verify_mode;
-    int depth    = con->ctx->verify_depth;
+    int errnum            = X509_STORE_CTX_get_error(ctx);
+    int errdepth          = X509_STORE_CTX_get_error_depth(ctx);
+    int verify            = con->ctx->verify_mode;
+    int depth             = con->ctx->verify_depth;
+    int ocsp_check_type   = con->ctx->no_ocsp_check;
+    int ocsp_soft_fail    = con->ctx->ocsp_soft_fail;
+    int ocsp_timeout      = con->ctx->ocsp_timeout;
+    int ocsp_verify_flags = con->ctx->ocsp_verify_flags;
 
-    if (verify == SSL_CVERIFY_UNSET ||
-        verify == SSL_CVERIFY_NONE)
+#if defined(SSL_OP_NO_TLSv1_3)
+    con->pha_state = PHA_COMPLETE;
+#endif
+
+    if (verify == SSL_CVERIFY_UNSET || verify == SSL_CVERIFY_NONE) {
         return 1;
+    }
 
-    if (SSL_VERIFY_ERROR_IS_OPTIONAL(errnum) &&
-        (verify == SSL_CVERIFY_OPTIONAL_NO_CA)) {
-        ok = 1;
+    if ((SSL_VERIFY_ERROR_IS_OPTIONAL(errnum) || errnum == X509_V_OK) && (verify == SSL_CVERIFY_OPTIONAL_NO_CA)) {
         SSL_set_verify_result(ssl, X509_V_OK);
+        // Skip OCSP checks since the CA is optional
+        return 1;
     }
 
     /*
@@ -331,27 +358,34 @@ int SSL_callback_SSL_verify(int ok, X509_STORE_CTX *ctx)
         X509_STORE_CTX_set_error(ctx, -1);
     }
 
-#ifdef HAVE_OCSP_STAPLING
+#ifdef HAVE_OCSP
     /* First perform OCSP validation if possible */
-    if (ok) {
-        /* If there was an optional verification error, it's not
-         * possible to perform OCSP validation since the issuer may be
-         * missing/untrusted.  Fail in that case.
-         */
-        if (SSL_VERIFY_ERROR_IS_OPTIONAL(errnum)) {
-            X509_STORE_CTX_set_error(ctx, X509_V_ERR_APPLICATION_VERIFICATION);
-            errnum = X509_V_ERR_APPLICATION_VERIFICATION;
-            ok = 0;
-        }
-        else {
-            int ocsp_response = ssl_verify_OCSP(ok, ctx);
-            if (ocsp_response == OCSP_STATUS_REVOKED) {
-                ok = 0 ;
-                errnum = X509_STORE_CTX_get_error(ctx);
+    if (ocsp_check_type == 0) {
+       if (ok) {
+            /* If there was an optional verification error, it's not
+             * possible to perform OCSP validation since the issuer may be
+             * missing/untrusted.  Fail in that case.
+             */
+            if (SSL_VERIFY_ERROR_IS_OPTIONAL(errnum)) {
+                X509_STORE_CTX_set_error(ctx, X509_V_ERR_APPLICATION_VERIFICATION);
+                errnum = X509_V_ERR_APPLICATION_VERIFICATION;
+                ok = 0;
             }
-            else if (ocsp_response == OCSP_STATUS_UNKNOWN) {
-                /* TODO: do nothing for time being */
-                ;
+            else {
+                int ocsp_response = ssl_verify_OCSP(ctx, ocsp_timeout, ocsp_verify_flags);
+                if (ocsp_response == OCSP_STATUS_REVOKED) {
+                    ok = 0 ;
+                    errnum = X509_STORE_CTX_get_error(ctx);
+                }
+                else if (ocsp_response == OCSP_STATUS_UNKNOWN) {
+                    errnum = X509_STORE_CTX_get_error(ctx);
+                    if (errnum != 0 && !(ocsp_soft_fail && errnum == X509_V_ERR_UNABLE_TO_GET_CRL)) {
+                        ok = 0;
+                    } else {
+                        // Clear the store error else the handshake will fail with the APR connector
+                        X509_STORE_CTX_set_error(ctx, 0);
+                    }
+                }
             }
         }
     }
@@ -386,12 +420,24 @@ int SSL_callback_SSL_verify(int ok, X509_STORE_CTX *ctx)
 void SSL_callback_handshake(const SSL *ssl, int where, int rc)
 {
     tcn_ssl_conn_t *con = (tcn_ssl_conn_t *)SSL_get_app_data(ssl);
+#ifdef HAVE_TLSV1_3
+    const SSL_SESSION *session = SSL_get_session(ssl);
+#endif
 
     /* Retrieve the conn_rec and the associated SSLConnRec. */
     if (con == NULL) {
         return;
     }
 
+#ifdef HAVE_TLSV1_3
+    /* TLS 1.3 does not use renegotiation so do not update the renegotiation
+     * state once we know we are using TLS 1.3. */
+    if (session != NULL) {
+        if (SSL_SESSION_get_protocol_version(session) == TLS1_3_VERSION) {
+            return;
+        }
+    }
+#endif
 
     /* If the reneg state is to reject renegotiations, check the SSL
      * state machine and move to ABORT if a Client Hello is being
@@ -405,18 +451,6 @@ void SSL_callback_handshake(const SSL *ssl, int where, int rc)
     else if ((where & SSL_CB_HANDSHAKE_DONE) && con->reneg_state == RENEG_INIT) {
         con->reneg_state = RENEG_REJECT;
     }
-
-}
-
-int SSL_callback_next_protos(SSL *ssl, const unsigned char **data,
-                             unsigned int *len, void *arg)
-{
-    tcn_ssl_ctxt_t *ssl_ctxt = arg;
-
-    *data = ssl_ctxt->next_proto_data;
-    *len = ssl_ctxt->next_proto_len;
-
-    return SSL_TLSEXT_ERR_OK;
 }
 
 /* The code here is inspired by nghttp2
@@ -477,22 +511,15 @@ int select_next_proto(SSL *ssl, const unsigned char **out, unsigned char *outlen
     return SSL_TLSEXT_ERR_NOACK;
 }
 
-int SSL_callback_select_next_proto(SSL *ssl, unsigned char **out, unsigned char *outlen,
-                         const unsigned char *in, unsigned int inlen,
-                         void *arg) {
-    tcn_ssl_ctxt_t *ssl_ctxt = arg;
-    return select_next_proto(ssl, (const unsigned char **) out, outlen, in, inlen, ssl_ctxt->next_proto_data, ssl_ctxt->next_proto_len, ssl_ctxt->next_selector_failure_behavior);
-}
-
 int SSL_callback_alpn_select_proto(SSL* ssl, const unsigned char **out, unsigned char *outlen,
         const unsigned char *in, unsigned int inlen, void *arg) {
     tcn_ssl_ctxt_t *ssl_ctxt = arg;
     return select_next_proto(ssl, out, outlen, in, inlen, ssl_ctxt->alpn_proto_data, ssl_ctxt->alpn_proto_len, ssl_ctxt->alpn_selector_failure_behavior);
 }
-#ifdef HAVE_OCSP_STAPLING
+#ifdef HAVE_OCSP
 
 /* Function that is used to do the OCSP verification */
-static int ssl_verify_OCSP(int ok, X509_STORE_CTX *ctx)
+static int ssl_verify_OCSP(X509_STORE_CTX *ctx, int timeout, int verifyFlags)
 {
     X509 *cert, *issuer;
     int r = OCSP_STATUS_UNKNOWN;
@@ -504,36 +531,32 @@ static int ssl_verify_OCSP(int ok, X509_STORE_CTX *ctx)
          * may yield NULL. Return early, but leave the ctx error as is. */
         return OCSP_STATUS_UNKNOWN;
     }
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
-    else if (cert->valid && X509_check_issued(cert,cert) == X509_V_OK) {
-#else
     /* No need to check cert->valid, because ssl_verify_OCSP() only
      * is called if OpenSSL already successfully verified the certificate
      * (parameter "ok" in SSL_callback_SSL_verify() must be true).
      */
     else if (X509_check_issued(cert,cert) == X509_V_OK) {
-#endif
         /* don't do OCSP checking for valid self-issued certs */
         X509_STORE_CTX_set_error(ctx, X509_V_OK);
         return OCSP_STATUS_UNKNOWN;
     }
 
     /* if we can't get the issuer, we cannot perform OCSP verification */
-    if (X509_STORE_CTX_get1_issuer(&issuer, ctx, cert) == 1 ) {
-        r = ssl_ocsp_request(cert, issuer);
-        if (r == OCSP_STATUS_REVOKED) {
+    issuer = X509_STORE_CTX_get0_current_issuer(ctx);
+    if (issuer != NULL) {
+        r = ssl_ocsp_request(cert, issuer, ctx, timeout, verifyFlags);
+        switch (r) {
+        case OCSP_STATUS_OK:
+            X509_STORE_CTX_set_error(ctx, X509_V_OK);
+            break;
+        case OCSP_STATUS_REVOKED:
             /* we set the error if we know that it is revoked */
             X509_STORE_CTX_set_error(ctx, X509_V_ERR_CERT_REVOKED);
+            break;
+        case OCSP_STATUS_UNKNOWN:
+            /* ssl_ocsp_request() sets the error correctly already. */
+            break;
         }
-        else {
-            /* else we return unknown */
-            r = OCSP_STATUS_UNKNOWN;
-        }
-        X509_free(issuer); /* It appears that we  should free issuer since
-                            * X509_STORE_CTX_get1_issuer() calls X509_OBJECT_up_ref_count()
-                            * on the issuer object (unline X509_STORE_CTX_get_current_cert()
-                            * that just returns the pointer
-                            */
     }
     return r;
 }
@@ -549,6 +572,58 @@ static void *apr_xrealloc(void *buf, size_t oldlen, size_t len, apr_pool_t *p)
     return newp;
 }
 
+/* Parses an ASN.1 length.
+ * On entry, asn1 points to the current tag.
+ * Updates the pointer to the ASN.1 structure to point to the start of the data.
+ * Returns 0 on success, 1 on failure.
+ */
+static int parse_asn1_length(unsigned char **asn1, int *len) {
+
+    /* Length immediately follows tag so increment before reading first (and
+     * possibly only) length byte.
+     */
+    (*asn1)++;
+
+    if (**asn1 & 0x80) {
+        // MSB set. Remaining bits are number of bytes used to store the length.
+        int i, l;
+
+        // How many bytes for this length?
+        i = **asn1 & 0x7F;
+
+        if (i == 0) {
+            /* This is the indefinite form of length. Since certificates use DER
+             * this should never happen and is therefore an error.
+             */
+            return 1;
+        }
+        if (i > 3) {
+            /* Three bytes for length gives a maximum of 16MB which should be
+             * far more than is required. (2 bytes is 64K which is probably more
+             * than enough but play safe.)
+             */
+            return 1;
+        }
+
+        // Most significant byte is first
+        l = 0;
+        while (i > 0) {
+            l <<= 8;
+            (*asn1)++;
+            l += **asn1;
+            i--;
+        }
+        *len = l;
+    } else {
+        // Single byte length
+        *len = **asn1;
+    }
+
+    (*asn1)++;
+
+    return 0;
+}
+
 /* parses the ocsp url and updates the ocsp_urls and nocsp_urls variables
    returns 0 on success, 1 on failure */
 static int parse_ocsp_url(unsigned char *asn1, char ***ocsp_urls,
@@ -558,11 +633,13 @@ static int parse_ocsp_url(unsigned char *asn1, char ***ocsp_urls,
     int len, err = 0, new_nocsp_urls;
 
     if (*asn1 == ASN1_STRING) {
-        len = *++asn1;
-        asn1++;
-        new_nocsp_urls = *nocsp_urls+1;
-        if ((new_ocsp_urls = apr_xrealloc(*ocsp_urls,*nocsp_urls, new_nocsp_urls, p)) == NULL)
-            err = 1;
+        err = parse_asn1_length(&asn1, &len);
+
+        if (!err) {
+            new_nocsp_urls = *nocsp_urls+1;
+            if ((new_ocsp_urls = apr_xrealloc(*ocsp_urls,*nocsp_urls, new_nocsp_urls, p)) == NULL)
+                err = 1;
+        }
         if (!err) {
             *ocsp_urls  = new_ocsp_urls;
             *nocsp_urls = new_nocsp_urls;
@@ -587,9 +664,9 @@ static int parse_ASN1_OID(unsigned char *asn1, char ***ocsp_urls, int *nocsp_url
     int len, err = 0 ;
     const unsigned char OCSP_OID[] = {0x2b, 0x06, 0x01, 0x05, 0x05, 0x07, 0x30, 0x01};
 
-    len = *++asn1;
-    asn1++;
-    if (memcmp(asn1, OCSP_OID, len) == 0) {
+    err = parse_asn1_length(&asn1, &len);
+
+    if (!err && len == 8 && memcmp(asn1, OCSP_OID, 8) == 0) {
         asn1+=len;
         err = parse_ocsp_url(asn1, ocsp_urls, nocsp_urls, p);
     }
@@ -611,13 +688,14 @@ static int parse_ASN1_Sequence(unsigned char *asn1, char ***ocsp_urls,
     while (!err && *asn1 != '\0') {
         switch(*asn1) {
             case ASN1_SEQUENCE:
-                len = *++asn1;
-                asn1++;
-                err = parse_ASN1_Sequence(asn1, ocsp_urls, nocsp_urls, p);
+                err = parse_asn1_length(&asn1, &len);
+                if (!err) {
+                    err = parse_ASN1_Sequence(asn1, ocsp_urls, nocsp_urls, p);
+                }
             break;
             case ASN1_OID:
                 err = parse_ASN1_OID(asn1,ocsp_urls,nocsp_urls, p);
-                return 0;
+                return err;
             break;
             default:
                 err = 1; /* we shouldn't have any errors */
@@ -643,33 +721,34 @@ static char **decode_OCSP_url(ASN1_OCTET_STRING *os, apr_pool_t *p)
     memcpy(ocsp_urls,os->data, len);
     ocsp_urls[len] = '\0';
 
-    if ((response = apr_pcalloc(p, sizeof(char *))) == NULL)
+    if ((response = apr_pcalloc(p, sizeof(char *))) == NULL) {
         return NULL;
-    if (parse_ASN1_Sequence(ocsp_urls, &response, &numofresponses, p))
+    }
+    if (parse_ASN1_Sequence(ocsp_urls, &response, &numofresponses, p) ||
+            numofresponses ==0) {
         response = NULL;
+    }
     return response;
 }
 
 
 /* stolen from openssl ocsp command */
-static int add_ocsp_cert(OCSP_REQUEST **req, X509 *cert, X509 *issuer,
-                         STACK_OF(OCSP_CERTID) *ids)
+static int add_ocsp_cert(OCSP_REQUEST *req, X509 *cert, X509 *issuer)
 {
     OCSP_CERTID *id;
 
     if (!issuer)
         return 0;
-    if (!*req)
-        *req = OCSP_REQUEST_new();
-    if (!*req)
-        return 0;
     id = OCSP_cert_to_id(NULL, cert, issuer);
-    if (!id || !sk_OCSP_CERTID_push(ids, id))
+    if (!id)
         return 0;
-    if (!OCSP_request_add0_id(*req, id))
+    if (!OCSP_request_add0_id(req, id)) {
+        OCSP_CERTID_free(id);
         return 0;
-    else
+    } else {
+        /* id will be freed by OCSP_REQUEST_free() */
         return 1;
+    }
 }
 
 
@@ -731,7 +810,6 @@ static int ocsp_send_req(apr_socket_t *sock, BIO *req)
     int len;
     char buf[TCN_BUFFER_SZ];
     apr_status_t rv;
-    int ok = 1;
 
     while ((len = BIO_read(req, buf, sizeof buf)) > 0) {
         char *wbuf = buf;
@@ -745,12 +823,11 @@ static int ocsp_send_req(apr_socket_t *sock, BIO *req)
         } while (rv == APR_SUCCESS && remain > 0);
 
         if (rv != APR_SUCCESS) {
-            apr_socket_close(sock);
-            ok = 0;
+            return 0;
         }
     }
 
-    return ok;
+    return 1;
 }
 
 
@@ -771,9 +848,6 @@ static OCSP_RESPONSE *parse_ocsp_resp(char *buf, int len)
 
     BIO_write(mem, buf, len);  /* write the buffer to the bio */
     if (BIO_gets(mem, tmpbuf, 512) <= 0) {
-#if OPENSSL_VERSION_NUMBER < 0x10100000L
-        OCSPerr(OCSP_F_OCSP_SENDREQ_BIO,OCSP_R_SERVER_RESPONSE_PARSE_ERROR);
-#endif
         goto err;
     }
     /* Parse the HTTP response. This will look like this:
@@ -838,19 +912,20 @@ err:
 
 /* Reads the response from the APR socket to a buffer, and parses the buffer to
    return the OCSP response  */
-#define ADDLEN 512
-static OCSP_RESPONSE *ocsp_get_resp(apr_socket_t *sock)
+#define BUFFER_SIZE 512
+#define OCSP_MAX_RESPONSE_SIZE 65536
+static OCSP_RESPONSE *ocsp_get_resp(apr_pool_t *mp, apr_socket_t *sock)
 {
     int buflen;
     apr_size_t totalread = 0;
     apr_size_t readlen;
-    char *buf, tmpbuf[ADDLEN];
+    char *buf, tmpbuf[BUFFER_SIZE];
     apr_status_t rv = APR_SUCCESS;
     apr_pool_t *p;
     OCSP_RESPONSE *resp;
 
-    apr_pool_create(&p, NULL);
-    buflen = ADDLEN;
+    apr_pool_create(&p, mp);
+    buflen = BUFFER_SIZE;
     buf = apr_palloc(p, buflen);
     if (buf == NULL) {
         apr_pool_destroy(p);
@@ -861,13 +936,16 @@ static OCSP_RESPONSE *ocsp_get_resp(apr_socket_t *sock)
         readlen = sizeof(tmpbuf);
         rv = apr_socket_recv(sock, tmpbuf, &readlen);
         if (rv == APR_SUCCESS) { /* if we have read something .. we can put it in the buffer*/
-            if ((totalread + readlen) >= buflen) {
-                buf = apr_xrealloc(buf, buflen, buflen + ADDLEN, p);
+            if ((totalread + readlen) > OCSP_MAX_RESPONSE_SIZE) {
+                apr_pool_destroy(p);
+                return NULL;
+            } else if ((totalread + readlen) >= buflen) {
+                buf = apr_xrealloc(buf, buflen, buflen * 2, p);
                 if (buf == NULL) {
                     apr_pool_destroy(p);
                     return NULL;
                 }
-                buflen += ADDLEN; /* if needed we enlarge the buffer */
+                buflen *= 2; /* if needed we enlarge the buffer */
             }
             memcpy(buf + totalread, tmpbuf, readlen); /* the copy to the buffer */
             totalread += readlen; /* update the total bytes read */
@@ -888,89 +966,134 @@ static OCSP_RESPONSE *ocsp_get_resp(apr_socket_t *sock)
     return resp;
 }
 
-/* Creates and OCSP request and returns the OCSP_RESPONSE */
-static OCSP_RESPONSE *get_ocsp_response(X509 *cert, X509 *issuer, char *url)
+/* Creates an OCSP request */
+static OCSP_REQUEST *get_ocsp_request(X509 *cert, X509 *issuer)
 {
-    OCSP_RESPONSE *ocsp_resp = NULL;
     OCSP_REQUEST *ocsp_req = NULL;
-    BIO *bio_req;
-    char *hostname, *path, *c_port;
-    int port, use_ssl;
-    STACK_OF(OCSP_CERTID) *ids = NULL;
-    int ok = 0;
-    apr_socket_t *apr_sock = NULL;
-    apr_pool_t *mp;
 
-    apr_pool_create(&mp, NULL);
-    ids = sk_OCSP_CERTID_new_null();
-
-    /* problem parsing the URL */
-    if (OCSP_parse_url(url,&hostname, &c_port, &path, &use_ssl) == 0 ) {
-        sk_OCSP_CERTID_free(ids);
-        return NULL;
-    }
-
-    /* Create the OCSP request */
-    if (sscanf(c_port, "%d", &port) != 1)
-        goto end;
     ocsp_req = OCSP_REQUEST_new();
     if (ocsp_req == NULL)
         return NULL;
-    if (add_ocsp_cert(&ocsp_req,cert,issuer,ids) == 0 )
-        goto free_req;
+
+    // Populate the request
+    if (add_ocsp_cert(ocsp_req, cert, issuer) == 0) {
+        OCSP_REQUEST_free(ocsp_req);
+        return NULL;
+    }
+
+    // Add a nonce to protect against replay attacks
+    OCSP_request_add1_nonce(ocsp_req, NULL, -1);
+
+    return ocsp_req;
+}
+
+/* Submits an OCSP request and returns the OCSP_RESPONSE */
+static OCSP_RESPONSE *get_ocsp_response(apr_pool_t *p, char *url, OCSP_REQUEST *ocsp_req, int timeout)
+{
+    OCSP_RESPONSE *ocsp_resp = NULL;
+    BIO *bio_req;
+    char *hostname, *path, *c_port;
+    int port, use_ssl;
+    int ok = 0;
+    apr_socket_t *apr_sock = NULL;
+    apr_pool_t *mp;
+#if OPENSSL_VERSION_NUMBER < 0x30000000L || defined(LIBRESSL_VERSION_NUMBER)
+    if (OCSP_parse_url(url, &hostname, &c_port, &path, &use_ssl) == 0)
+#else
+    if (OSSL_HTTP_parse_url(url, &use_ssl, NULL, &hostname, &c_port, NULL, &path, NULL, NULL) == 0)
+#endif
+        goto end;
+
+    if (sscanf(c_port, "%d", &port) != 1)
+        goto end;
 
     /* create the BIO with the request to send */
     bio_req = serialize_request(ocsp_req, hostname, port, path);
     if (bio_req == NULL) {
-        goto free_req;
+        goto end;
     }
 
+    apr_pool_create(&mp, p);
     apr_sock = make_socket(hostname, port, mp);
     if (apr_sock == NULL) {
-        ocsp_resp = NULL;
         goto free_bio;
     }
 
+    apr_socket_timeout_set(apr_sock, timeout);
+
     ok = ocsp_send_req(apr_sock, bio_req);
-    if (ok)
-        ocsp_resp = ocsp_get_resp(apr_sock);
+    if (ok) {
+        ocsp_resp = ocsp_get_resp(mp, apr_sock);
+    }
+    apr_socket_close(apr_sock);
 
 free_bio:
     BIO_free(bio_req);
-
-free_req:
-    if(apr_sock && ok) /* if ok == 0 we have already closed the socket */
-        apr_socket_close(apr_sock);
-
     apr_pool_destroy(mp);
 
-    sk_OCSP_CERTID_free(ids);
-    OCSP_REQUEST_free(ocsp_req);
-
 end:
+    OPENSSL_free(hostname);
+    OPENSSL_free(c_port);
+    OPENSSL_free(path);
+
     return ocsp_resp;
 }
 
 /* Process the OCSP_RESPONSE and returns the corresponding
-   answert according to the status.
+   answer according to the status.
 */
-static int process_ocsp_response(OCSP_RESPONSE *ocsp_resp)
+static int process_ocsp_response(OCSP_REQUEST *ocsp_req, OCSP_RESPONSE *ocsp_resp, X509 *cert, X509 *issuer,
+        X509_STORE_CTX *ctx, int verifyFlags)
 {
     int r, o = V_OCSP_CERTSTATUS_UNKNOWN, i;
     OCSP_BASICRESP *bs;
     OCSP_SINGLERESP *ss;
+    OCSP_CERTID *certid;
+    ASN1_GENERALIZEDTIME *thisupd;
+    ASN1_GENERALIZEDTIME *nextupd;
+    const STACK_OF(X509) *certStack;
 
     r = OCSP_response_status(ocsp_resp);
 
     if (r != OCSP_RESPONSE_STATUS_SUCCESSFUL) {
-        OCSP_RESPONSE_free(ocsp_resp);
         return OCSP_STATUS_UNKNOWN;
     }
+
     bs = OCSP_response_get1_basic(ocsp_resp);
+    if (OCSP_check_nonce(ocsp_req, bs) == 0) {
+        X509_STORE_CTX_set_error(ctx, X509_V_ERR_OCSP_RESP_INVALID);
+        o = OCSP_STATUS_UNKNOWN;
+        goto clean_bs;
+    }
 
-    ss = OCSP_resp_get0(bs,0); /* we know we have only 1 request */
+    certStack = OCSP_resp_get0_certs(bs);
+    // Cast to non-const pointer is OK here since OCSP_basic_verify does not modify the provided certs
+    if (OCSP_basic_verify(bs, (STACK_OF(X509) *)certStack, X509_STORE_CTX_get0_store(ctx), verifyFlags) <= 0) {
+        X509_STORE_CTX_set_error(ctx, X509_V_ERR_OCSP_SIGNATURE_FAILURE);
+        o = OCSP_STATUS_UNKNOWN;
+        goto clean_bs;
+    }
+
+    certid = OCSP_cert_to_id(NULL, cert, issuer);
+    if (certid == NULL) {
+        X509_STORE_CTX_set_error(ctx, X509_V_ERR_OCSP_RESP_INVALID);
+        o = OCSP_STATUS_UNKNOWN;
+        goto clean_bs;
+    }
+
+    ss = OCSP_resp_get0(bs, OCSP_resp_find(bs, certid, -1)); /* find by serial number and get the matching response */
+    i = OCSP_single_get0_status(ss, NULL, NULL, &thisupd, &nextupd);
+    if (OCSP_check_validity(thisupd, nextupd, OCSP_MAX_SKEW, -1) <= 0) {
+        X509_STORE_CTX_set_error(ctx, X509_V_ERR_OCSP_NOT_YET_VALID);
+        o = OCSP_STATUS_UNKNOWN;
+        goto clean_certid;
+    }
+    if (OCSP_check_validity(thisupd, nextupd, OCSP_MAX_SKEW, OCSP_MAX_SKEW) <= 0) {
+        X509_STORE_CTX_set_error(ctx, X509_V_ERR_OCSP_HAS_EXPIRED);
+        o = OCSP_STATUS_UNKNOWN;
+        goto clean_certid;
+    }
 
-    i = OCSP_single_get0_status(ss, NULL, NULL, NULL, NULL);
     if (i == V_OCSP_CERTSTATUS_GOOD)
         o =  OCSP_STATUS_OK;
     else if (i == V_OCSP_CERTSTATUS_REVOKED)
@@ -978,15 +1101,18 @@ static int process_ocsp_response(OCSP_RESPONSE *ocsp_resp)
     else if (i == V_OCSP_CERTSTATUS_UNKNOWN)
         o = OCSP_STATUS_UNKNOWN;
 
-    /* we clean up */
-    OCSP_RESPONSE_free(ocsp_resp);
+clean_certid:
+    OCSP_CERTID_free(certid);
+clean_bs:
+    OCSP_BASICRESP_free(bs);
     return o;
 }
 
-static int ssl_ocsp_request(X509 *cert, X509 *issuer)
+static int ssl_ocsp_request(X509 *cert, X509 *issuer, X509_STORE_CTX *ctx, int timeout, int verifyFlags)
 {
     char **ocsp_urls = NULL;
     int nid;
+    int rv = OCSP_STATUS_UNKNOWN;
     X509_EXTENSION *ext;
     ASN1_OCTET_STRING *os;
     apr_pool_t *p;
@@ -1005,19 +1131,35 @@ static int ssl_ocsp_request(X509 *cert, X509 *issuer)
     /* if we find the extensions and we can parse it check
        the ocsp status. Otherwise, return OCSP_STATUS_UNKNOWN */
     if (ocsp_urls != NULL) {
-        OCSP_RESPONSE *resp;
+        OCSP_REQUEST *req;
+        OCSP_RESPONSE *resp = NULL;
         /* for the time being just check for the fist response .. a better
            approach is to iterate for all the possible ocsp urls */
-        resp = get_ocsp_response(cert, issuer, ocsp_urls[0]);
+        req = get_ocsp_request(cert, issuer);
+        if (req != NULL) {
+            resp = get_ocsp_response(p, ocsp_urls[0], req, timeout);
+            if (resp != NULL) {
+                rv = process_ocsp_response(req, resp, cert, issuer, ctx, verifyFlags);
+            } else {
+                /* Unable to send request / receive response. */
+                X509_STORE_CTX_set_error(ctx, X509_V_ERR_UNABLE_TO_GET_CRL);
+            }
+        } else {
+            /* correct error code for application errors? */
+            X509_STORE_CTX_set_error(ctx, X509_V_ERR_APPLICATION_VERIFICATION);
+        }
+
+        if (req != NULL) {
+            OCSP_REQUEST_free(req);
+        }
 
         if (resp != NULL) {
-            apr_pool_destroy(p);
-            return process_ocsp_response(resp);
+            OCSP_RESPONSE_free(resp);
         }
     }
     apr_pool_destroy(p);
-    return OCSP_STATUS_UNKNOWN;
+    return rv;
 }
 
-#endif /* HAVE_OCSP_STAPLING */
+#endif /* HAVE_OCSP */
 #endif /* HAVE_OPENSSL  */
